-- This is a simple mod specifically for use together with the inserter
-- throughput savegame. It finds all the combinator holding test results
-- and exports their readings to a csv file.

local inserter_order = {
  ["burner-inserter"] = 1,
  ["inserter"] = 2,
  ["long-handed-inserter"] = 3,
  ["fast-inserter"] = 4,
  ["filter-inserter"] = 5,
  ["stack-inserter"] = 6,
  ["filter-stack-inserter"] = 7,
}

local belt_order = {
  ["transport-belt"] = 1,
  ["fast-transport-belt"] = 2,
  ["express-transport-belt"] = 3,
}

function entity_gps_tag(entity)
  return "[gps=" .. entity.position.x .. "," .. entity.position.y .. "]"
end

function export_results()
  local sfc = game.player.surface
  local acs = sfc.find_entities_filtered({type="arithmetic-combinator"})

  -- Find the test settings for each global green wire. This is done through
  -- a dummy 1/0 arithmetic combinator which has the setting signals on its
  -- input and the green wire on its output
  local test_settings = {}
  for _, ac in pairs(acs) do
    local cb = ac.get_control_behavior()
    local params = cb.parameters.parameters
    if params ~= nil and
       params.first_signal == nil and
       params.first_constant == 1 and
       params.second_signal == nil and
       params.second_constant == 0
    then
      local cn_settings = cb.get_circuit_network(
        defines.wire_type.green, defines.circuit_connector_id.combinator_input)
      local cn_global_green = cb.get_circuit_network(
        defines.wire_type.green, defines.circuit_connector_id.combinator_output)
      if cn_settings ~= nil and #cn_settings.signals > 0 and
         cn_global_green ~= nil then
        local inserter_type, belt_type
        for _, s in pairs(cn_settings.signals) do
          if s.signal.type == "item" then
            if inserter_order[s.signal.name] ~= nil then
              inserter_type = s.signal.name
            end
            if belt_order[s.signal.name] ~= nil then
              belt_type = s.signal.name
            end
          end
        end
        local signal_o =
          cn_global_green.get_signal({type="virtual", name="signal-O"})
        if inserter_type == nil or belt_type == nil then
          game.player.print(entity_gps_tag(ac) ..
            " Ignored testsuite that doesn't specify inserter and belt")
          test_settings[cn_global_green.network_id] = {ignored = true}
        elseif signal_o ~= nil and signal_o > 0 then
          game.player.print(entity_gps_tag(ac) ..
            " Ignored currently running testsuite")
          test_settings[cn_global_green.network_id] = {ignored = true}
        else
          local inserter_dirs = ""
          if cn_settings.get_signal({type="virtual", name="signal-1"}) == 1 then
            inserter_dirs = inserter_dirs .. "S"
          end
          if cn_settings.get_signal({type="virtual", name="signal-2"}) == 1 then
            inserter_dirs = inserter_dirs .. "W"
          end
          if cn_settings.get_signal({type="virtual", name="signal-3"}) == 1 then
            inserter_dirs = inserter_dirs .. "N"
          end
          if cn_settings.get_signal({type="virtual", name="signal-4"}) == 1 then
            inserter_dirs = inserter_dirs .. "E"
          end
          test_settings[cn_global_green.network_id] = {
            ac = ac,
            inserter_type = inserter_type,
            belt_type = belt_type,
            I = cn_settings.get_signal({type="virtual", name="signal-I"}),
            W = cn_settings.get_signal({type="virtual", name="signal-W"}),
            A = cn_settings.get_signal({type="virtual", name="signal-A"}),
            B = cn_settings.get_signal({type="virtual", name="signal-B"}),
            inserter_dirs = inserter_dirs,
          }
        end
      end
    end
  end

  -- Read all the data export combinators.
  local all_data = {}
  for _, ac in pairs(acs) do
    local cb = ac.get_control_behavior()
    local params = cb.parameters.parameters
    if params ~= nil and
       params.first_signal ~= nil and
       params.first_signal.type == "virtual" and
       params.first_signal.name == "signal-V" and
       params.operation == "+" and
       params.second_signal == nil and
       params.second_constant == 1 and
       params.output_signal ~= nil and
       params.output_signal.type == "virtual" and
       params.output_signal.name == "signal-V"
    then
      local cn_red = cb.get_circuit_network(
        defines.wire_type.red, defines.circuit_connector_id.combinator_input)
      local cn_green = cb.get_circuit_network(
        defines.wire_type.green, defines.circuit_connector_id.combinator_input)
      if cn_green == nil or test_settings[cn_green.network_id] == nil then
        game.player.print(entity_gps_tag(ac) ..
          " Found data export combinator without known green wire connected.")
      elseif test_settings[cn_green.network_id].ignored then
        -- Ignored testsuite.
      else
        local testsuite_data = all_data[cn_green.network_id]
        if testsuite_data == nil then
          testsuite_data = {}
          all_data[cn_green.network_id] = testsuite_data
        end
        testsuite_data[#testsuite_data + 1] = {
          ac = ac,
          V = cn_red.get_signal({type="virtual", name="signal-V"}),
          H = cn_red.get_signal({type="virtual", name="signal-H"}),
          L = cn_red.get_signal({type="virtual", name="signal-L"}),
          T = cn_red.get_signal({type="virtual", name="signal-T"}),
          N = cn_red.get_signal({type="virtual", name="signal-N"}),
          O = cn_red.get_signal({type="virtual", name="signal-O"}),
        }
      end
    end
  end

  function testsuite_data_comp(a, b)
    if a.V == b.V then
      game.player.print("Duplicate V in " .. entity_gps_tag(a.ac) .. " and " ..
                        entity_gps_tag(b.ac))
    end
    return a.V < b.V
  end

  local testsuite_rows = {}
  for id,testsuite_data in pairs(all_data) do
    local settings = test_settings[id]
    local rows = {}
    table.sort(testsuite_data, testsuite_data_comp)
    for i = 1, #testsuite_data do
      local values = testsuite_data[i]
      local cols = {
        settings.inserter_type or "",
        settings.I or 0,
        settings.belt_type or "",
        values.V or 0,
        values.H or 0,
        values.L or 0,
        values.T or 0,
        values.N or 0,
        values.O or 0,
        settings.O or 0,
        settings.W or 0,
        settings.A or 0,
        settings.B or 0,
        settings.inserter_dirs or "",
      }
      rows[#rows + 1] = table.concat(cols, ",")
    end
    testsuite_rows[#testsuite_rows + 1] = {settings, table.concat(rows, "\n")}
  end

  function testsuite_comp(a, b)
    -- NB sorts in reverse, starting with the generally fastest throughputs.
    if a == b then
      -- table.sort can apparently compare items with themselves. :\
      return false
    end
    local sa = a[1]
    local sb = b[1]
    if inserter_order[sa.inserter_type] ~= inserter_order[sb.inserter_type] then
      return inserter_order[sa.inserter_type] > inserter_order[sb.inserter_type]
    end
    if sa.I ~= sb.I then
      return sa.I > sb.I
    end
    if belt_order[sa.belt_type] ~= belt_order[sb.belt_type] then
      return belt_order[sa.belt_type] > belt_order[sb.belt_type]
    end
    game.player.print(
      "Found two active testsuites with same setup " ..
      sa.inserter_type .. "," .. sa.I .. "," .. sa.belt_type .. ": " ..
      entity_gps_tag(sa.ac) .. " and " .. entity_gps_tag(sb.ac))
    return false
  end
  table.sort(testsuite_rows, testsuite_comp)

  local file_data = {}
  for i = 1, #testsuite_rows do
    file_data[#file_data + 1] = testsuite_rows[i][2]
  end
  game.write_file("results-tick-" .. game.tick .. ".csv",
                  table.concat(file_data, "\n"))
end

commands.add_command(
  "export-results",
  "Exports all measurement results to csv files",
  export_results)

function find_stuck_rigs()
  local sfc = game.player.surface
  local acs = sfc.find_entities_filtered({type="arithmetic-combinator"})

  local stuck_global_green_wires = {}
  for _, ac in pairs(acs) do
    local cb = ac.get_control_behavior()
    local params = cb.parameters.parameters
    if params ~= nil and
       params.first_signal == nil and
       params.first_constant == 1 and
       params.second_signal == nil and
       params.second_constant == 0
    then
      local cn_settings = cb.get_circuit_network(
        defines.wire_type.green, defines.circuit_connector_id.combinator_input)
      local cn_global_green = cb.get_circuit_network(
        defines.wire_type.green, defines.circuit_connector_id.combinator_output)
      if cn_settings ~= nil and #cn_settings.signals > 0 and
         cn_global_green ~= nil then
        local inserter_type, belt_type
        for _, s in pairs(cn_settings.signals) do
          if s.signal.type == "item" then
            if inserter_order[s.signal.name] ~= nil then
              inserter_type = s.signal.name
            end
            if belt_order[s.signal.name] ~= nil then
              belt_type = s.signal.name
            end
          end
        end
        local signal_d =
          cn_global_green.get_signal({type="virtual", name="signal-D"})
        if inserter_type == nil or belt_type == nil then
          game.player.print(entity_gps_tag(ac) ..
            " Ignored testsuite that doesn't specify inserter and belt")
        elseif signal_d ~= nil and signal_d > 5000 then
          -- Ought to check what the alert limit is for stuckness, rather than
          -- hardcoding 5000.
          stuck_global_green_wires[cn_global_green.network_id] = true
        end
      end
    end
  end

  local dcs = sfc.find_entities_filtered({type="decider-combinator"})
  for _, dc in pairs(dcs) do
    local cb = dc.get_control_behavior()
    local cn = cb.get_circuit_network(
      defines.wire_type.green, defines.circuit_connector_id.combinator_output)
    local params = cb.parameters.parameters
    if cn ~= nil and
       stuck_global_green_wires[cn.network_id] and
       params ~= nil and
       params.first_signal ~= nil and
       params.first_signal.type == "virtual" and
       params.first_signal.name == "signal-T" and
       params.comparator == "=" and
       params.second_signal == nil and
       params.constant == 0 and
       params.output_signal ~= nil and
       params.output_signal.type == "virtual" and
       params.output_signal.name == "signal-U"
    then
      local signal_u =
        cb.get_signal_last_tick({type="virtual", name="signal-U"})
      if signal_u ~= nil and signal_u > 0 then
        game.player.print(entity_gps_tag(dc) .. " Combinator sends U > 0")
      end
    end
  end
end

commands.add_command(
  "find-stuck-rigs",
  "Finds the combinators sending U=1 on the green wire when the " ..
  "\"long test run\" alert is firing",
  find_stuck_rigs)
